# frozen_string_literal: true

class Array
  # Returns a new array created by sorting +self+, using merge sort algorithm.
  #
  # Comparisons for the sort will be done using the <=> operator or using an optional code block.
  #
  # The block must implement a comparison between +a+ and +b+ and return an integer less than 0 when +b+ follows +a+,
  # +0+ when +a+ and +b+ are equivalent, or an integer greater than 0 when +a+ follows +b+.
  #
  # The result is guaranteed to be stable. When the comparison of two elements returns +0+, the order of the elements
  # will be preserved.
  #
  # @return [Array] the sorted array
  def merge_sort(&block)
    return dup if length <= 1
    divided_parts = merge_sort_divide
    part_a_sorted = divided_parts[0].merge_sort
    part_b_sorted = divided_parts[1].merge_sort
    merge_sort_merge part_a_sorted, part_b_sorted, &block
  end

  # Sorts +self+ in place, using merge sort algorithm.
  #
  # Comparisons for the sort will be done using the <=> operator or using an optional code block.
  #
  # The block must implement a comparison between +a+ and +b+ and return an integer less than 0 when +b+ follows +a+,
  # +0+ when +a+ and +b+ are equivalent, or an integer greater than 0 when +a+ follows +b+.
  #
  # The result is guaranteed to be stable. When the comparison of two elements returns +0+, the order of the elements
  # will be preserved.
  #
  # @return [Array] +self+
  def merge_sort!(&block)
    become_clone_of merge_sort(&block)
  end

  # Returns a new array created by sorting +self+ with merge sort algorithm, using a set of keys generated by mapping
  # the values in self through the given block.
  #
  # The result is guaranteed to be stable. When the comparison of two elements returns +0+, the order of the elements
  # will be preserved.
  #
  # If no block is given, an Enumerator is returned instead.
  #
  # @return [Array] if a block is given, the sorted array
  # @return [Enumerator] if no block is given, an Enumerator
  def merge_sort_by(&_block)
    if block_given?
      merge_sort do |a, b|
        yield(a) <=> yield(b)
      end
    else
      to_enum :merge_sort_by
    end
  end

  # Sorts +self+ in place with merge sort algorithm, using a set of keys generated by mapping the values in self through
  # the given block.
  #
  # The result is guaranteed to be stable. When the comparison of two elements returns +0+, the order of the elements
  # will be preserved.
  #
  # If no block is given, an Enumerator is returned instead.
  #
  # @return [Array] if a block is given, the sorted array
  # @return [Enumerator] if no block is given, an Enumerator
  def merge_sort_by!(&block)
    if block_given?
      become_clone_of merge_sort_by(&block)
    else
      to_enum :merge_sort_by!
    end
  end

  private

  # @private
  def merge_sort_divide
    mid = length / 2
    part_a = self[0, mid]
    part_b = self[mid..-1]
    [part_a, part_b]
  end

  # @private
  def merge_sort_merge(part_a, part_b, &block)
    result = []
    while part_a.length.positive? && part_b.length.positive?
      result << (sort_compare(part_a.first, part_b.first, &block).positive? ? part_b : part_a).shift
    end
    result + part_a + part_b
  end
end
